;; $Id: heute.lisp,v 1.18 2006/01/06 14:19:18 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(defpackage "HEUTE"
  (:use "COMMON-LISP"
	#+cmu   "PCL"
	#+sbcl  "SB-MOP"
	#+clisp "CLOS")
  (:export
   "ASSERT-CONDITION"
   "ASSERT-EQUAL"
   "ASSERT-FALSE"
   "ASSERT-NOT-CONDITION"
   "ASSERT-NOT-EQUAL"
   "ASSERT-TRUE"
   "CURRENT-RESULT"
   "FAIL-IF"
   "FAIL-IF-CONDITION"
   "FAIL-IF-EQUAL"
   "FAIL-IF-NOT"
   "FAIL-IF-NOT-CONDITION"
   "FAIL-IF-NOT-EQUAL"   
   "FINAL-ACTION"
   "GET-CURRENT-TEST-FUN"
   "INTERNAL-FAIL-IF"
   "NUMBER-OF-SIBLINGS"
   "PARENT-SUITE"
   "REGISTER-GUI"
   "RESULTS"
   "RUN-ALL-UNIT-TESTS"
   "RUN-ALL-UNIT-TESTS*"
   "RUN-SUITE"
   "RUN-TEST"
   "SIBLING-INDEX"
   "STATUS"
   "SUITE-LEVEL"
   "TAG"
   "TEST-FUN"
   "TEST-FUNS"
   "TEST-NAME"
   "TESTCASE"
   "TEXT"
   "WHY-FAILED"
   ))


(in-package heute)

#|
(setf (fdefinition 'ensure-class)
      (fdefinition
       #+cmu 'pcl::ensure-class
       #+clisp 'clos::ensure-class
       #+sbcl  'sb-mop::ensure-class))

(setf (fdefinition 'class-direct-subclasses)
      (fdefinition
       #+cmu 'pcl::class-direct-subclasses
       #+clisp 'clos::class-direct-subclasses
       #+sbcl 'sb-mop::class-direct-subclasses))

(setf (fdefinition 'class-precedence-list)
      (fdefinition
       #+cmu 'pcl::class-precedence-list
       #+clisp 'clos::class-precedence-list
       #+sbcl 'sb-mop::class-precedence-list))

|#

(defgeneric internal-fail-if ( test-obj expression
					&key tag text action-on-fail))
(defgeneric current-result (test))
(defgeneric run-suite (test))
(defgeneric run-test (test test-fun))
(defgeneric status (obj))
(defgeneric test-name (obj))
(defgeneric test-funs (obj))
(defgeneric final-action ( testcase result text))
(defgeneric test-suites ( testcase))
(defgeneric suite-level (testcase))
(defgeneric sibling-index (testcase))
(defgeneric number-of-siblings (testcase))
(defgeneric tag (testcase))
(defgeneric test-fun (testcase))
(defgeneric text (testcase))

(defclass heute-base ()
  ())

(defvar *gui-mixins* nil)
(setf *gui-mixins* nil)
(defmacro register-gui ( gui-class)
  `(progn
     (pushnew ',gui-class *gui-mixins*)
     (format t "guis:  ~A~%"  *gui-mixins*)
     (ensure-class 'heute-gui-mixin
		   :direct-superclasses *gui-mixins*)
     (make-instance 'testcase)))
     
;; This class is intended to be redefined
;; once per gui that needs to be added.
;; Please use the function register-gui to register
;; your gui class(es)
(defclass heute-gui-mixin nil nil)

(defclass testcase (heute-base heute-gui-mixin)
  (
   ;; You should give each subclass of TESTCASE
   ;; explictly a new value of TEST-NAME via the
   ;; slot :INITFORM
   ( test-name
     :accessor test-name
     :initform "Top Level Test")
   ( test-funs
     :documentation "List of strings nameing methods defined on the class"
     :accessor test-funs
     :initform nil)
   ;; The following slots are not inteded to be mentioned
   ;; when defining subclasses.
   ( level
     :accessor suite-level
     :initarg :suite-level
     :initform 0)
   ( sibling-index
     :accessor sibling-index
     :initform 0
     :initarg :sibling-index)
   ( test-suites
     :accessor test-suites
     :initform nil)
   ( parent-suite
     :accessor parent-suite
     :initarg :parent
     :initform nil)
   ( results
     :documentation "list of instances of class TEST-RESULT"
     :initform nil
     :accessor results)))

(defmethod initialize-instance :after (( testcase testcase) &key &allow-other-keys)
  (setf (test-suites testcase)
	(loop for subclass in (class-direct-subclasses (class-of testcase))
	      for index from 0
	      collect (make-instance subclass
				     :parent testcase
				     :sibling-index index
				     :suite-level (1+ (suite-level testcase)))))
  (when (equal (test-name testcase)
	       (test-name (parent-suite testcase)))
    (format t "class ~A has the same name (~A) its parent ~A~%"
	    (class-of testcase)
	    (test-name testcase)
	    (class-of (parent-suite testcase)))))

(defmethod test-name (( obj t))
  nil)

(defmethod number-of-siblings (( testcase testcase))
  (length (test-suites (parent-suite testcase))))

(defgeneric why-failed ( testcase stream))

(defmethod why-failed (( testcase testcase) stream)
  (fresh-line)
  (dolist (result (results testcase))
    (when (eql :fail (status result))
      (format stream "why?  test-fun=~A tag=~A status=~A text=~A~%"
	      (test-fun result) (tag result) (status result) (text result))))
  (dolist (suite (test-suites testcase))
    (when (eql :fail (status suite))
      (format stream "why?  name=~A status=~A~%"
	      (test-name suite) (status suite))))
  (finish-output))     


(defmethod status (( testcase testcase))
  (cond
    ((some #'(lambda (result) (eql :fail (status result)))
	   (results testcase))

     :fail)
    ((some #'(lambda (suite) (eql :fail (status suite)))
	   (test-suites testcase))
     :fail)
    (t
     :pass)))

(defvar *test-statuses* '(:not-yet-run
			  :pass
			  :fail
			  :running
			  :omitted))

(defclass test-result (heute-base)
  (( testcase
     :accessor testcase
     :initarg :testcase)
   ( test-fun
     :accessor test-fun
     :initform nil
     :initarg :test-fun)
   ( tag
     :initform nil
     :accessor tag
     :initarg :tag)
   ( status
     :accessor status
     :initarg :status
     :initform :not-yet-run)
   ( text
     :accessor text
     :initarg :text
     :initform nil)
   ( action-on-fail
     :accessor action-on-fail
     :initarg :action-on-fail
     :initform nil)
   ( log
     :initform nil)))

(defmethod final-action (testcase ( result test-result) text)
  (declare (ignore testcase result text))
  nil)

(defclass test-result-pass (test-result)
  (( status
     :initform :pass)))

(defmethod status (( result test-result-pass))
  :pass)

(defclass test-result-fail (test-result)
  (( status
     :initform :fail)))

(defmethod status (( result test-result-fail))
  :fail)

(defclass test-result-error (test-result-fail)
  (( condition-obj
     :documentation "holds the second argument returned from ignore-errors"
     :accessor condition-obj
     :initarg :condition-obj
     :initform nil))
  (:documentation "if a test fails because of an error, this result class is used"))

(defgeneric set-condition ( result error))
(defmethod set-condition (( result test-result) (error (eql nil)))
  nil)
(defmethod set-condition (( result test-result-error) error)
  (setf (condition-obj result) error))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; a graphical UI may add before or after methods
;; to any of these generic functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; A subclass of testcase may add before or after
;; methods to any of run-suite or run-test
;; to implement the setup and mopup functionality.


(defmethod current-result (( test testcase))
  (car (results test)))

(let (current-test-fun)

  (defun get-current-test-fun ()
    current-test-fun)

  (defmethod run-test (( testcase testcase) test-fun)
    (setf current-test-fun test-fun)
    (catch 'testing-function
      ;; The test-fun is likely to make several calls to
      ;;  fail-if-* which are macros around internal-fail-if.
      ;;  If the :action-on-fail keywork argument to internal-fail-if
      ;;  is :abort-test, then internal-fail-if throws to this catch.
      (multiple-value-bind ( value error)
	  (ignore-errors (funcall test-fun testcase))
	(declare (ignore value))
	(when error
	  (push (make-instance 'test-result-error
			 :condition-obj error
			 :testcase testcase
			 :test-fun test-fun
			 :text "failed outside an assert")
		(results testcase))))))
	  
  (defmethod internal-fail-if (( testcase testcase) value-closure
			       &key tag text action-on-fail)
    (let (( result (make-instance 'test-result
				  :status :running
				  :testcase testcase
				  ;; enclosed variable current-test-fun
				  :test-fun current-test-fun
				  :action-on-fail action-on-fail
				  :tag tag)))

      (push result (results testcase))

      (multiple-value-bind ( value error)
	  (ignore-errors (funcall value-closure))
	(finish-output)
	(change-class result
		      (cond
			(error 'test-result-error)
			(value 'test-result-fail)
			(t     'test-result-pass)))
	(set-condition result error))

      (final-action testcase result text)))

  (defmethod final-action ( testcase ( result test-result-fail) text)
    (declare (ignore testcase))
    (setf (text result) (funcall text))
    (when (eql :abort-test (action-on-fail result))
      (throw 'testing-function t))))

(defmethod run-suite (( testcase testcase))
  (setf (results testcase) nil)
  (dolist (suite (test-suites testcase))
    (run-suite suite))
  (dolist (test-fun (test-funs testcase))
    (run-test testcase test-fun))
  testcase)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; testing API
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro fail-if ( testcase expression &key tag text action-on-fail)
  `(internal-fail-if ,testcase
    #'(lambda () ,expression)
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

;; same as fail-if
(defmacro assert-false ( testcase expression &key tag text action-on-fail)
  `(internal-fail-if ,testcase
    #'(lambda () ,expression)
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro fail-if-not ( testcase expression &key tag text action-on-fail)
  `(internal-fail-if ,testcase
    #'(lambda () (not ,expression))
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro assert-true ( testcase expression &key tag text action-on-fail)
  `(internal-fail-if ,testcase
    #'(lambda () (not ,expression))
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro fail-if-equal ( testcase lhs rhs
			 &key tag text action-on-fail (test ''eql))
  `(internal-fail-if ,testcase
    #'(lambda () (funcall ,test ,lhs ,rhs))
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro assert-not-equal ( testcase lhs rhs
			    &key tag text action-on-fail (test ''eql))
  ;; assert-not-equal is fail-if-equal
  `(internal-fail-if ,testcase
    #'(lambda () (funcall ,test ,lhs ,rhs))
    :tag ,tag
    :text (lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro fail-if-not-equal ( testcase lhs rhs
			     &key tag text action-on-fail (test ''eql))
  `(internal-fail-if ,testcase
    #'(lambda () (not (funcall ,test ,lhs ,rhs)))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro assert-equal ( testcase lhs rhs
			&key tag text action-on-fail (test ''eql))
  ;; assert-equal is fail-if-not-equal
  `(internal-fail-if ,testcase
    #'(lambda () (not (funcall ,test ,lhs ,rhs)))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))


(defmacro fail-if-condition ( testcase condition expression
			     &key tag text action-on-fail )
  (assert (symbolp condition))
  `(internal-fail-if ,testcase
    #'(lambda () (block try
		   (handler-bind
		       ((,condition
			 (lambda (ign) (declare (ignore ign))
				 (return-from try t))))
		     ;; ignore the return value because we are only
		     ;; testing for the condition
		     (prog1 nil ,expression))))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro assert-not-condition ( testcase condition expression
				&key tag text action-on-fail )
  (assert (symbolp condition))
  `(internal-fail-if ,testcase
    #'(lambda () (block try
		   (handler-bind
		       ((,condition
			 (lambda (ign) (declare (ignore ign))
				 (return-from try t))))
		     ;; ignore the return value because we are only
		     ;; testing for the condition
		     (prog1 nil ,expression))))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro fail-if-not-condition ( testcase condition expression
				 &key tag text action-on-fail )
  (assert (symbolp condition))
  `(internal-fail-if ,testcase
    #'(lambda () (block try
		 (handler-bind
		     ((,condition
		       (lambda (ign) (declare (ignore ign))
			       (return-from try nil))))
		   ;; ignore the return value because we are only
		   ;; testing for the condition
		   (prog1 t ,expression))))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))

(defmacro assert-condition ( testcase condition value
			    &key tag text action-on-fail )
  (assert (symbolp condition))
  `(internal-fail-if ,testcase
    #'(lambda () (block try
		 (handler-bind
		     ((,condition
		       (lambda (ign) (declare (ignore ign))
			       (return-from try nil))))
		   ;; ignore the return value because we are only
		   ;; testing for the condition
		   (prog1 t ,value))))
    :tag ,tag
    :text #'(lambda () ,text)
    :action-on-fail ,action-on-fail))


(defgeneric run-all-unit-tests (obj))

(defmethod run-all-unit-tests (( obj (eql t)))
  (run-all-unit-tests 'testcase))

(defmethod run-all-unit-tests (( class-name symbol))
  (run-all-unit-tests (make-instance class-name)))

(defmethod run-all-unit-tests (( testcase testcase))
  (run-suite testcase))

(defmacro run-all-unit-tests* (&optional (arg t))
  `(run-all-unit-tests ,arg))
